/*
 *	LICENSE:
 *	Copyright (c) 2000-2001
 *		Charles D. Mercadal, Michigan 49001.
 *			charleschuck@mac.com
 *	    All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with or without
 *	modification, are permitted provided that the following conditions
 *	are met:
 *	1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer as
 *	the first lines of the source code file unmodified.
 *	2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *	3. Neither name of the Charles Mercadal nor the names of its contributors may be
 *	used to endorse or promote products derived from this software without
 *	specific prior written permission.
 *	
 *	
 *	THIS SOFTWARE IS PROVIDED BY CHARLES MERCADAL "AS IS" AND ANY EXPRESS OR
 *	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL CHARLES MERCADAL BE LIABLE FOR ANY DIRECT, INDIRECT,
 *	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	@(#) checks.c - 0.3.3 - Charles Mercadal - 8 November 2001
 */

/*	I suggest setting tabstops to 4 for best reading	*/

#ifndef CHECKS_H
#include "checks.h"
#endif	/*	CHECKS_H	*/

int main( int argc, char *argv[] )
{
    pref mainPref;				/*	the main pref struct	*/
    info *dirinfo;				/*	Array of directory info	*/
    register int counter;		/*	counter variable		*/
    char usingp = 0;			/*	using -p option		*/
    int ch;
    extern char	*optarg;
    extern int optind;
    
    mainPref.minOrSec = 's';		/*	default value		*/
    mainPref.sleepTime = 60;		/*	If none is given	*/
    mainPref.nrm = 0;			/*	alert when files removed*/
    
    while( ( ch = getopt( argc, argv, "m:s:pn" ) ) != EOF )
    {
        switch( ch )
        {
            case 'm':
                mainPref.minOrSec = 'm';
                mainPref.sleepTime = atoi( optarg );
                break;
            case 's':
                mainPref.minOrSec = 's';
                mainPref.sleepTime = atoi( optarg );
                break;
            case 'p':
                usingp = 1;
                break;
            case 'n':
                mainPref.nrm = 1;
                break;
            case '?':
            default:
                displayHelp();
                break;
        }
    }
    argc -= optind;
    argv += optind;
    
    if( usingp )
    {
        mainPref.numberOfDirs = getFileLength();
        
        if( !( dirinfo = (info *)calloc( mainPref.numberOfDirs, sizeof( info ) ) ) )
            err( eMallocFoobarExit, BadMalloc );
            
        parseChecksrc( dirinfo );
    }
    else
    {
        if( argc <= 0 )		/*	no directory given	*/
            displayHelp();
        
        if( !( dirinfo = (info *)calloc( ( mainPref.numberOfDirs = argc ), sizeof( info ) ) ) )
            err( eMallocFoobarExit, BadMalloc );
            
        for( counter = 0; counter < argc; counter++ )
        {
            dirinfo[ counter ].presentValue = -1;
            
            if( !( dirinfo[ counter ].directory = strdup( argv[ counter ] ) ) )
                err( eMallocFoobarExit, BadMalloc );
        }
    }
    
    /*	Run printDirectories after initPrefs	*/
    initPrefs( &mainPref );
    printDirectories( dirinfo, mainPref.numberOfDirs );
    
    for( ; ; )
    {	
        for( ( counter = mainPref.numberOfDirs - 1 ); counter >= 0; counter-- )
            doEventLoop( &mainPref, &dirinfo[ counter ] );

        if( usingp )
            flushLastVal( dirinfo, mainPref.numberOfDirs );

        sleep( mainPref.sleepTime );
    }
    
    /*	How the hell could we ever get here?	*/
    return 1;
    
}
    
/*
 *	displayHelp() will inform uninformed users how 
 *	to properly use the arguments available to checks.
 */
void displayHelp( void )
{
	(void)fprintf( stderr, "Usage:\tchecks [-n] [-ms delay time] [[-p] | [directories...]]\n-m:\tRequires delay in minutes\n-s:\tRequires delay in seconds\n-n:\tDo not alert if files are removed from directory\n-p:\tUse ~/.checksrc\nComments?  mercadal@khons.diablonet.net\n" );
	exit( eBadUserExit );
}

/*
 *	initPrefs will set up the pref struct with the values that
 *	the other functions in the program expect, as well as tell the 
 *	user that we are really actually checking the given directories
 *	for them now.
 */
void initPrefs( pref *toAdjust )
{
    if( toAdjust->minOrSec == 'm' )
    {
        if( toAdjust->sleepTime < 1 )	/*	if luser enters < 1	*/
            toAdjust->sleepTime = 1;
        
        (void)printf( "\nWill check the following folders approximately every %d minutes:\n", toAdjust->sleepTime );
        toAdjust->sleepTime = toAdjust->sleepTime * 60;
        toAdjust->minOrSec = 's';	/*	sleep() takes values in seconds	*/
    }
    else	/*	already in seconds form	*/
    {
        if( ( toAdjust->sleepTime < 10 ) && ( getuid() != 0 ) )
            toAdjust->sleepTime = 10;	/*	Don't let lusers grind disk too hard	*/
            
        (void)printf( "\nWill check the following folders approximately every %d seconds:\n", toAdjust->sleepTime );
    }
}

/*
 *	printDirectories just prints out the directories that 
 *	are actually being checked.  Run after initPrefs
 */
void printDirectories( const info *dirArray, register int counter )
{
    while( counter > 0 )
    {
        --counter;

        (void)puts( dirArray[ counter ].directory );
    }
}

/*
 *	doEventLoop is a poorly named function, as we really 
 *	aren't getting any events, and the loop is external.  However, 
 *	this function actually does the job of checking if the 
 *	size of our directory has changed, and will print out
 *	the message that it has (if it has)
 */
void doEventLoop( const pref *ourPrefs, info *dirinfo )
{
    FILE *pipe;
    char duop[ strlen( Du ) + strlen( dirinfo->directory ) + 1 ];
    long int previous = dirinfo->presentValue;
    
    (void)sprintf( duop, "%s%s", Du, dirinfo->directory );
    
    if( !( pipe = popen( duop, "r" ) ) )
        return;		/*	maybe we can read from pipe next run	*/
        
    (void)fscanf( pipe, "%ld", &dirinfo->presentValue );
    
    pclose( pipe );
    
    if( previous != -1 )
    {
        if( ( ( previous > dirinfo->presentValue ) && !( ourPrefs->nrm ) ) || previous < dirinfo->presentValue )
            (void)printf( "\nFolder %s has changed\a\n", dirinfo->directory );
        
    }
    
}

/*
 *	parseChecksrc will do it's darndest to read the .checksrc in
 *	the user's home folder, extract directory names, and, if available, 
 *	extract the size of the directory at last run through
 */
void parseChecksrc( info *dinfo )
{
    FILE *infile;		/*	parse this thing		*/
    char *stringloc;		/*	pointer to read char *		*/
    register int counter;		/*	general counter var		*/
    register int masterCounter = 0;	/*	main counter for input		*/
	char success;
    char *prefLoc;

    prefFilePath( &prefLoc );
    
    if( !( infile = fopen( prefLoc, "r" ) ) )
        err( eBadFile, BadFileIO );
    
    free( prefLoc );

    while( ( stringloc = getline( infile ) ) )
    {
		/*	first character is white space or comment character	*/
        if( isspace( *stringloc ) || (*stringloc == '#') )
            continue;

        for( counter = 0; stringloc[ counter ] != '\0'; counter++ )
        {
            if( stringloc[ counter ] == '\t' )
            {
				success = 1;
                if( !( dinfo[ masterCounter ].directory = (char *)malloc( counter ) ) )
                    err( eMallocFoobarExit, BadMalloc );

				/*	Copy string stringloc up to tabstop	*/
                strncpy( dinfo[ masterCounter ].directory, stringloc, counter );
                dinfo[ masterCounter ].directory[ counter ] = '\0';
                
                dinfo[ masterCounter ].presentValue = atoi( &stringloc[ counter + 1 ] );
            }
        }

        if( !success )	/*	new directory entry	*/
        {
            if( !( dinfo[ masterCounter ].directory = strdup( stringloc ) ) )
                err( eMallocFoobarExit, BadMalloc );
            
            dinfo[ masterCounter ].presentValue = -1;
        }

        /* Get ready for next loop */
        free( stringloc );
        success = 0;
        ++masterCounter;
    }

    fclose( infile );

}

/*
 *	flushLastVal will dump the values from the given info struct
 *	into a checks readable file at .checksrc
 */
void flushLastVal( const info *toFlush, int howManyDirectories )
{
    sigset_t allSignals;
    FILE *outfile;
    char *prefLoc;	/*	preference file location	*/

    prefFilePath( &prefLoc );
    sigfillset( &allSignals );		/*	allSignals initialized with all signals	*/

    if( !( outfile = fopen( prefLoc, "w" ) ) )
        err( eBadFile, BadFileIO );
    
    free( prefLoc );
    
	/*
	 *	Careful!  We stop receiving proper signals here
	 */
    sigprocmask( SIG_BLOCK, &allSignals, NULL );
    while( howManyDirectories > 0 )
    {
        --howManyDirectories;
        
        (void)fprintf(	outfile,
                    "%s\t%ld\n",
                    toFlush[ howManyDirectories ].directory,
                    toFlush[ howManyDirectories ].presentValue );
    }
    
    fclose( outfile );
    sigprocmask( SIG_UNBLOCK, &allSignals, NULL);
	/*
	 *	...and here's where proper signals return
	 */
}

/*
 *	This function will return the length of a file, well, really
 *	how many parsable lines there are in ~/.checksrc
 */
int getFileLength( void )
{
    register int counter = 0;
    FILE *infile;
    char *firstChar;		/* char * from fgetln */
    char *prefLoc;

    prefFilePath( &prefLoc );
        
    if( !( infile = fopen( prefLoc, "r" ) ) )	/*	Damn...no .checksrc	*/
        err( eBadFile, BadFileIO );
        
    free( prefLoc );		/*	after use, free what prefFilePath returns	*/

    while( ( firstChar = getline( infile ) ) )
    {
        if( isspace( *firstChar ) || ( *firstChar == '#' ) )
            continue;
                
        ++counter;
    }

    fclose( infile );

    return counter;
}

/*
 *	prefFilePath gets the path of .checksrc and sets toSet to that value
 */
void prefFilePath( char **toSet )
{
    if( !( *toSet = (char *)malloc( strlen( getenv( "HOME" ) ) + strlen( PrefFile ) + 1 ) ) )
        err( eMallocFoobarExit, BadMalloc );

    (void)sprintf( *toSet, "%s%s", getenv( "HOME" ), PrefFile );
}

/*
 *	getline gets a line from the file at fp and returns it as 
 *	a null-terminated string.  What more could you want?
 */
char *getline( register FILE *fp )
{
	register unsigned char *pos;
	char *ret;
	register size_t length;

	/*
	 *	Do we have any more input?  If we haven't,
	 *	we should probably return some sort of value 
	 *	stating that
	 */
	if( fp->_r <= 0 && __srefill( fp ) )
		return NULL;
                
	/*
	 *	So, we search for a newline.  If one is found, 
	 *	we figure the length of the string in relation
	 *	to fp->_p.  If not, we figure it in relation to 
	 *	fp->_r (what's left in the file)  For more on this
	 *	disgusting topic of the FILE type, see
	 *	/usr/include/stdio.h
	 */
	if( ( pos = memchr( (void *)fp->_p, '\n', fp->_r ) ) )
		length = ( ++pos - fp->_p );
	else
		length = ( fp->_r + 1 );

	if( !( ret = (char *)malloc( length * sizeof( char ) ) ) )
		err( 1, NULL );
	ret = memcpy( ret, (void *)fp->_p, length );
	ret[ length - 1 ] = '\0';
	fp->_p = pos;
	fp->_r -= length;
	return ret;
}
